import argparse
from pathlib import Path

from rstb import ResourceSizeTable, util
from . import builder, unbuilder


def rstb_to_json():
    parser = argparse.ArgumentParser(description="Converts a binary RSTB file to JSON.")
    parser.add_argument(
        "-b",
        "--be",
        action="store_true",
        help="Read the RSTB as big endian for Wii U, otherwise little endian for Switch",
    )
    parser.add_argument("rstb", help="Path to a binary RSTB file", nargs="?")
    parser.add_argument("-o", "--output", help="Path to output JSON file", nargs="?")
    args = parser.parse_args()

    in_file = Path(args.rstb).resolve()
    table = util.read_rstb(str(in_file), be=args.be)
    output = (
        Path(args.output).resolve()
        if args.output
        else in_file.with_suffix(".rsizetable.json")
    )
    unbuilder.rstb_to_json(table, output, set())


def json_to_rstb():
    parser = argparse.ArgumentParser(description="Converts a JSON RSTB file to binary.")
    parser.add_argument(
        "-b",
        "--be",
        action="store_true",
        help="Write the RSTB as big endian for Wii U, otherwise little endian for Switch",
    )
    parser.add_argument("json", help="Path to a JSON RSTB file", nargs="?")
    parser.add_argument("-o", "--output", help="Path to output RSTB binary", nargs="?")
    args = parser.parse_args()

    in_file = Path(args.json).resolve()
    table = builder.load_rstb(be=args.be, file=in_file)
    output = (
        Path(args.output).resolve()
        if args.output
        else in_file.with_suffix("").with_suffix(".srsizetable")
    )
    util.write_rstb(table, str(output), be=args.be)



# pylint: disable=missing-docstring
from pathlib import Path
from typing import Union

from oead import yaz0


def guess_bfres_size(file: Union[Path, bytes], be: bool, name: str = "") -> int:
    if isinstance(file, bytes):
        if file[0:4] == b"Yaz0":
            real_size = yaz0.get_header(memoryview(file[0:16])).uncompressed_size
        else:
            real_size = len(file)
        del file
    else:
        if file.suffix.startswith(".s"):
            with file.open("rb") as f:
                real_size = yaz0.get_header(f.read(16)).uncompressed_size
        else:
            real_size = file.stat().st_size
    real_size = int(real_size * 1.05)
    if name == "":
        if isinstance(file, Path):
            name = file.name
        else:
            raise ValueError("BFRES name must not be blank if passing file as bytes.")
    value: int
    if be:
        if ".Tex" in name:
            if real_size < 100:
                value = real_size * 9
            elif 100 < real_size <= 2000:
                value = real_size * 7
            elif 2000 < real_size <= 3000:
                value = real_size * 5
            elif 3000 < real_size <= 4000:
                value = real_size * 4
            elif 4000 < real_size <= 8500:
                value = real_size * 3
            elif 8500 < real_size <= 12000:
                value = real_size * 2
            elif 12000 < real_size <= 17000:
                value = real_size * 1.75
            elif 17000 < real_size <= 30000:
                value = real_size * 1.5
            elif 30000 < real_size <= 45000:
                value = real_size * 1.3
            elif 45000 < real_size <= 100000:
                value = real_size * 1.2
            elif 100000 < real_size <= 150000:
                value = real_size * 1.1
            elif 150000 < real_size <= 200000:
                value = real_size * 1.07
            elif 200000 < real_size <= 250000:
                value = real_size * 1.045
            elif 250000 < real_size <= 300000:
                value = real_size * 1.035
            elif 300000 < real_size <= 600000:
                value = real_size * 1.03
            elif 600000 < real_size <= 1000000:
                value = real_size * 1.015
            elif 1000000 < real_size <= 1800000:
                value = real_size * 1.009
            elif 1800000 < real_size <= 4500000:
                value = real_size * 1.005
            elif 4500000 < real_size <= 6000000:
                value = real_size * 1.002
            else:
                value = real_size * 1.0015
        else:
            if real_size < 500:
                value = real_size * 7
            elif 500 < real_size <= 750:
                value = real_size * 4
            elif 750 < real_size <= 2000:
                value = real_size * 3
            elif 2000 < real_size <= 400000:
                value = real_size * 1.75
            elif 400000 < real_size <= 600000:
                value = real_size * 1.7
            elif 600000 < real_size <= 1500000:
                value = real_size * 1.6
            elif 1500000 < real_size <= 3000000:
                value = real_size * 1.5
            else:
                value = real_size * 1.25
    else:
        if ".Tex" in name:
            if 50000 < real_size:
                value = real_size * 1.2
            elif 30000 < real_size:
                value = real_size * 1.3
            elif 10000 < real_size:
                value = real_size * 1.5
            else:
                value = real_size * 2
        else:
            if 4000000 < real_size:
                value = real_size * 1.5
            elif 3000000 < real_size:
                value = real_size * 1.667
            elif 2000000 < real_size:
                value = real_size * 2.5
            elif 800000 < real_size:
                value = real_size * 3.15
            elif 100000 < real_size:
                value = real_size * 3.5
            elif 50000 < real_size:
                value = real_size * 3.66
            elif 2500 < real_size:
                value = real_size * 4.25
            elif 1250 < real_size:
                value = real_size * 6
            else:
                value = real_size * 9.5
    return int(value)


def guess_aamp_size(file: Union[Path, bytes], be: bool, ext: str = "") -> int:
    if isinstance(file, bytes):
        real_size = len(file)
        del file
    else:
        if file.suffix.startswith(".s"):
            with file.open("rb") as f:
                real_size = yaz0.get_header(f.read(16)).uncompressed_size
        else:
            real_size = file.stat().st_size
    real_size = int(real_size * 1.05)
    if ext == "":
        if isinstance(file, Path):
            ext = file.suffix
        else:
            raise ValueError(
                "AAMP extension must not be blank if passing file as bytes."
            )
    ext = ext.replace(".s", ".")
    value: int
    if ext == ".baiprog":
        if real_size <= 380:
            value = real_size * 7
        elif 380 < real_size <= 400:
            value = real_size * 6
        elif 400 < real_size <= 450:
            value = real_size * 5.5
        elif 450 < real_size <= 600:
            value = real_size * 5
        elif 600 < real_size <= 1000:
            value = real_size * 4
        elif 1000 < real_size <= 1750:
            value = real_size * 3.5
        else:
            value = real_size * 3
    elif ext == ".bgparamlist":
        if real_size <= 100:
            value = real_size * 20
        elif 100 < real_size <= 150:
            value = real_size * 12
        elif 150 < real_size <= 250:
            value = real_size * 10
        elif 250 < real_size <= 350:
            value = real_size * 8
        elif 350 < real_size <= 450:
            value = real_size * 7
        else:
            value = real_size * 6
    elif ext == ".bdrop":
        if real_size < 200:
            value = real_size * 8.5
        elif 200 < real_size <= 250:
            value = real_size * 7
        elif 250 < real_size <= 350:
            value = real_size * 6
        elif 350 < real_size <= 450:
            value = real_size * 5.25
        elif 450 < real_size <= 850:
            value = real_size * 4.5
        else:
            value = real_size * 4
    elif ext == ".bxml":
        if real_size < 350:
            value = real_size * 6
        elif 350 < real_size <= 450:
            value = real_size * 5
        elif 450 < real_size <= 550:
            value = real_size * 4.5
        elif 550 < real_size <= 650:
            value = real_size * 4
        elif 650 < real_size <= 800:
            value = real_size * 3.5
        else:
            value = real_size * 3
    elif ext == ".brecipe":
        if real_size < 100:
            value = real_size * 12.5
        elif 100 < real_size <= 160:
            value = real_size * 8.5
        elif 160 < real_size <= 200:
            value = real_size * 7.5
        elif 200 < real_size <= 215:
            value = real_size * 7
        else:
            value = real_size * 6.5
    elif ext == ".bshop":
        if real_size < 200:
            value = real_size * 7.25
        elif 200 < real_size <= 400:
            value = real_size * 6
        elif 400 < real_size <= 500:
            value = real_size * 5
        else:
            value = real_size * 4.05
    elif ext == ".bas":
        real_size = real_size * 1.05
        if real_size < 100:
            value = real_size * 20
        elif 100 < real_size <= 200:
            value = real_size * 12.5
        elif 200 < real_size <= 300:
            value = real_size * 10
        elif 300 < real_size <= 600:
            value = real_size * 8
        elif 600 < real_size <= 1500:
            value = real_size * 6
        elif 1500 < real_size <= 2000:
            value = real_size * 5.5
        elif 2000 < real_size <= 15000:
            value = real_size * 5
        else:
            value = real_size * 4.5
    elif ext == ".baslist":
        if real_size < 100:
            value = real_size * 15
        elif 100 < real_size <= 200:
            value = real_size * 10
        elif 200 < real_size <= 300:
            value = real_size * 8
        elif 300 < real_size <= 500:
            value = real_size * 6
        elif 500 < real_size <= 800:
            value = real_size * 5
        elif 800 < real_size <= 4000:
            value = real_size * 4
        else:
            value = real_size * 3.5
    elif ext == ".bdmgparam":
        value = (((-0.0018 * real_size) + 6.6273) * real_size) + 500
    else:
        value = 0
    if not be:
        value = value * 1.5
    return int(value)



# pylint: disable=invalid-name,bare-except,missing-docstring,no-name-in-module,bad-continuation
import json
from datetime import datetime
from functools import partial
from multiprocessing import Pool, cpu_count
from os import system
from pathlib import Path
from shutil import rmtree
from typing import Union
from zlib import crc32

import oead
from oead import aamp
from oead.yaz0 import decompress # pylint: disable=import-error
import pymsyt
from rstb import ResourceSizeTable
from rstb.util import read_rstb

from . import AAMP_EXTS, BYML_EXTS, SARC_EXTS, STOCK_FILES, get_canon_name

HANDLED = {"ResourceSizeTable.product.srsizetable", "ActorInfo.product.sbyml"}


def _if_unyaz(data: bytes) -> bytes:
    return data if data[0:4] != b"Yaz0" else decompress(data)


def _unbuild_file(f: Path, out: Path, content: str, mod: Path, verbose: bool) -> set:
    of = out / f.relative_to(mod)
    if not of.parent.exists():
        of.parent.mkdir(parents=True, exist_ok=True)
    names = set()
    canon = get_canon_name(f.relative_to(mod))
    if canon:
        names.add(canon)
    if f.name in HANDLED:
        pass
    elif f.suffix in AAMP_EXTS:
        of.with_suffix(f"{f.suffix}.yml").write_bytes(_aamp_to_yml(f.read_bytes()))
    elif f.suffix in BYML_EXTS:
        of.with_suffix(f"{f.suffix}.yml").write_bytes(_byml_to_yml(f.read_bytes()))
    elif f.suffix in SARC_EXTS:
        with f.open("rb") as file:
            data = file.read()
            if data[0:4] == b"Yaz0":
                data = decompress(data)
            if data[0:4] != b"SARC":
                return names
            s = oead.Sarc(data)
        if "bactorpack" in f.suffix:
            names.update(_unbuild_actorpack(s, out / content))
        else:
            names.update(_unbuild_sarc(s, of))
        del s
    else:
        of.write_bytes(f.read_bytes())
    if verbose:
        print(f"Unbuilt {f.relative_to(mod).as_posix()}")
    return names


def rstb_to_json(rstb: ResourceSizeTable, output: Path, names: set):
    hash_map = {crc32(h.encode("utf8")): h for h in STOCK_FILES}
    hash_map.update({crc32(name.encode("utf8")): name for name in names})

    def hash_to_name(crc: int) -> Union[str, int]:
        return hash_map[crc] if crc in hash_map else crc

    output.parent.mkdir(parents=True, exist_ok=True)
    output.write_text(
        json.dumps(
            {
                "hash_map": {hash_to_name(k): v for k, v in rstb.crc32_map.items()},
                "name_map": rstb.name_map,
            },
            ensure_ascii=False,
            indent=2,
        )
    )


def _unbuild_rstb(content: Path, be: bool, out: Path, mod: Path, names: set):
    f: Path = content / "System" / "Resource" / "ResourceSizeTable.product.srsizetable"
    table = read_rstb(str(f), be=be)
    rstb_to_json(table, out / f.relative_to(mod).with_suffix(".json"), names)


def _unbuild_actorinfo(mod: Path, content: str, out: Path):
    file = mod / content / "Actor" / "ActorInfo.product.sbyml"
    actor_info = oead.byml.from_binary(decompress(file.read_bytes()))
    actor_info_dir = out / content / "Actor" / "ActorInfo"
    actor_info_dir.mkdir(parents=True, exist_ok=True)
    for actor in actor_info["Actors"]:
        output = actor_info_dir / f"{actor['name']}.info.yml"
        output.write_text(oead.byml.to_text(actor), encoding="utf-8")


def _unbuild_actorpack(s: oead.Sarc, output: Path):
    output.mkdir(parents=True, exist_ok=True)
    for f in {f for f in s.get_files() if "Actor" in f.name}:
        out = (output / f.name).with_suffix(f"{Path(f.name).suffix}.yml")
        out.parent.mkdir(parents=True, exist_ok=True)
        if f.data[0:4] == b"AAMP":
            out.write_bytes(_aamp_to_yml(f.data))
        elif f.data[0:2] in [b"BY", b"YB"]:
            out.write_bytes(_byml_to_yml(f.data))
    for f in {
        f for f in s.get_files() if "Physics" in f.name and "Actor" not in f.name
    }:
        out = output / f.name
        out.parent.mkdir(parents=True, exist_ok=True)
        out.write_bytes(f.data)
    return {f.name for f in s.get_files()}


def _unbuild_sarc(s: oead.Sarc, output: Path, skip_actorpack: bool = False):
    SKIP_SARCS = {
        "tera_resource.Cafe_Cafe_GX2.release.ssarc",
        "tera_resource.Nin_NX_NVN.release.ssarc",
    }

    output.mkdir(parents=True, exist_ok=True)
    if any(f.name.startswith("/") for f in s.get_files()):
        (output / ".slash").write_bytes(b"")

    names = set()
    for sarc_file in s.get_files():
        sf = sarc_file.name
        osf = output / sf
        names.add(sf.replace(".s", "."))
        if sf.startswith("/"):
            osf = output / sf[1:]
        osf.parent.mkdir(parents=True, exist_ok=True)
        ext = osf.suffix
        if ext in SARC_EXTS:
            if osf.name in SKIP_SARCS:
                osf.write_bytes(sarc_file.data)
                continue
            try:
                ss = oead.Sarc(_if_unyaz(sarc_file.data))
                if (
                    "bactorpack" in ext
                    and output.stem == "TitleBG"
                    and not skip_actorpack
                ):
                    names.update(_unbuild_actorpack(ss, output.parent.parent))
                else:
                    names.update(_unbuild_sarc(ss, osf))
                del ss
            except ValueError:
                osf.write_bytes(b"")
        elif ext in AAMP_EXTS:
            if osf.with_suffix(f"{osf.suffix}.yml").exists():
                continue
            osf.with_suffix(f"{osf.suffix}.yml").write_bytes(
                _aamp_to_yml(sarc_file.data)
            )
        elif ext in BYML_EXTS:
            osf.with_suffix(f"{osf.suffix}.yml").write_bytes(
                _byml_to_yml(sarc_file.data)
            )
        else:
            osf.write_bytes(sarc_file.data)

    if "Msg_" in output.name:
        pymsyt.export(output, output)
        rmtree(output)
        output.with_suffix("").rename(output)
    if output.suffix in {".ssarc", ".sarc"}:
        (output / ".align").write_text(str(s.guess_min_alignment()))
    return names


def _byml_to_yml(file_bytes: bytes) -> bytes:
    if file_bytes[0:4] == b"Yaz0":
        file_bytes = decompress(file_bytes)
    return oead.byml.to_text(oead.byml.from_binary(file_bytes)).encode("utf8")


def _aamp_to_yml(file_bytes: bytes) -> bytes:
    if file_bytes[0:4] == b"Yaz0":
        file_bytes = decompress(file_bytes)
    return aamp.ParameterIO.from_binary(file_bytes).to_text().encode("utf8")


def unbuild_mod(args) -> None:
    mod = Path(args.directory)
    if not any(
        d.exists()
        for d in {
            mod / "content",
            mod / "aoc",
            mod / "01007EF00011E000/romfs",
            mod / "01007EF00011F001/romfs",
        }
    ):
        print("The specified directory is not valid: no base game or DLC folder found")
        exit(1)
    out = mod.with_name(f"{mod.name}_unbuilt") if not args.output else Path(args.output)
    if out.exists():
        rmtree(out, True)
    be = (mod / "content").exists() or (mod / "aoc").exists()
    content = "content" if be else "01007EF00011E000/romfs"

    print("Analying files...")
    files = {f for f in mod.rglob("**/*") if f.is_file()}
    t = min(len(files), cpu_count())
    names = set()

    print("Unbuilding...")
    if args.single or t < 2:
        for f in files:
            names.update(_unbuild_file(f, out, content, mod, args.verbose))
    else:
        with Pool(processes=t) as p:
            result = p.map(
                partial(
                    _unbuild_file,
                    mod=mod,
                    content=content,
                    out=out,
                    verbose=args.verbose,
                ),
                files,
            )
            for r in result:
                if r:
                    names.update(r)

    print("Unpacking actor info...")
    if (mod / content / "Actor" / "ActorInfo.product.sbyml").exists():
        _unbuild_actorinfo(mod, content, out)

    print("Dumping RSTB...")
    if (
        mod / content / "System" / "Resource" / "ResourceSizeTable.product.srsizetable"
    ).exists():
        _unbuild_rstb(mod / content, be, out, mod, names)

    (out / ".done").write_text(str(datetime.now().timestamp()))
    try:
        system(f'attrib +h "{str(out / ".done")}"')
    except:
        pass

    print("Unbuilding complete")



# pylint: disable=bad-continuation,unsupported-assignment-operation
import argparse
from multiprocessing import set_start_method
from pathlib import Path
import oead
from . import AAMP_EXTS, BYML_EXTS, SARC_EXTS
from .unbuilder import _unbuild_sarc as unsarc


def unbuild_sarc() -> None:
    set_start_method("spawn", True)
    parser = argparse.ArgumentParser(
        description="Unbuild a single SARC file completely"
    )
    parser.add_argument("sarc", help="SARC archive to unbuild")
    parser.add_argument(
        "--output",
        "-O",
        help="Output folder for unbuilt SARC,\
        defaults to file name w/o extension",
    )
    args = parser.parse_args()

    try:
        file: Path = Path(args.sarc)
        data = file.read_bytes()
        if data[0:4] == b"Yaz0":
            data = oead.yaz0.decompress(data)
        unsarc(
            oead.Sarc(data),
            Path(args.output) if args.output else file.with_suffix(""),
            True,
        )
    except (FileNotFoundError, oead.InvalidDataError, OSError) as err:
        print(err)
        return


def build_sarc() -> None:
    set_start_method("spawn", True)
    parser = argparse.ArgumentParser(
        description="Build a SARC file from a single source folder"
    )
    parser.add_argument("source", help="Source folder for SARC")
    parser.add_argument(
        "output",
        help='Path to output SARC file, will auto compress if\
        extension starts with ".s"',
    )
    parser.add_argument(
        "--be", "-B", help="Use big endian where applicable", action="store_true"
    )
    parser.add_argument(
        "--verbose", "-V", help="Provide more detailed output", action="store_true"
    )
    args = parser.parse_args()

    source = Path(args.source)
    output = Path(args.output)

    def build_yaml(file: Path) -> bytes:
        real_file = file.with_suffix("")
        data: bytes
        if real_file.suffix in BYML_EXTS:
            data = oead.byml.to_binary(
                oead.byml.from_text(file.read_text("utf-8")), args.be
            )
        elif real_file.suffix in AAMP_EXTS:
            data = oead.aamp.ParameterIO.from_text(file.read_text("utf-8")).to_binary()
        else:
            raise TypeError("Can only build AAMP or BYML files from YAML")
        if real_file.suffix.startswith(".s"):
            data = oead.yaz0.compress(data)
        return data

    yml_table = {file: build_yaml(file) for file in source.rglob("**/*.yml")}
    all_files = {
        f
        for f in source.rglob("**/*")
        if f.is_file() or (f.is_dir() and f.suffix in SARC_EXTS)
    }
    nest_sarcs = {}

    for nest in sorted(
        {d for d in all_files if d.suffix in SARC_EXTS and d.is_dir()} | {source},
        key=lambda x: len(x.parts),
        reverse=True,
    ):
        if args.verbose:
            print(f"Building {nest.name}...")
        sarc = oead.SarcWriter(
            oead.Endianness.Big if args.be else oead.Endianness.Little
        )
        for file in all_files.copy():
            try:
                rel_path = file.relative_to(nest)
                assert file != nest
            except (ValueError, AssertionError):
                continue
            store_path = (
                rel_path.as_posix()
                if rel_path.suffix != ".yml"
                else rel_path.with_suffix("").as_posix()
            )
            if args.verbose:
                print(f"  Adding {store_path} to {nest.name}")
            sarc.files[store_path] = yml_table.pop(
                file, nest_sarcs.pop(file, None) or file.read_bytes()
            )
            all_files.remove(file)

        if nest.suffix.startswith(".s") and nest.suffix != ".sarc":
            sarc_data = oead.yaz0.compress(sarc.write()[1])
        else:
            sarc_data = sarc.write()[1]
        nest_sarcs[nest] = sarc_data
        if args.verbose:
            print(f"Finished building {nest.name}")

    output.write_bytes(nest_sarcs[source])


if __name__ == "__main__":
    build_sarc()
